;Animated LED Duck, 34 LEDs + speaker
;Padauk PMS152-S14/S16 (OTP) or PFS154-S14/S16 (Flash)
;Assembler: asl <http://john.ccac.rwth-aachen.de:8000/as/>
;
;PORTS
;  PA0   = LEDs 1 (anodes  1~ 6, cathodes     7, 13, 19, 25, 31)
;  PA3   = LEDs 4 (anodes 19~24, cathodes 3,  9, 15,     28, 34)
;  PA4   = LEDs 5 (anodes 25~30, cathodes 4, 10, 16, 22,     35)
;  PA5   = button (to ground, internal pull-up)
;  PA6   = LEDs 6 (anodes 31~34, cathodes 5, 11, 17, 23, 29    )
;  PA7   = LEDs 7 (              cathodes 6, 12, 18, 24, 30, 36)
;  PB0   = EEPROM SCL
;  PB1   = LEDs 2 (anodes  7~12, cathodes 1,     14, 20, 26, 32)
;  PB2   = LEDs 3 (anodes 13~18, cathodes 2,  8,     21, 27, 33)
;  PB3   = [unused/unavailable on -S14]
;  PB4   = [unused/unavailable on -S14]
;  PB5   = Speaker PWM signal 1 [LPG0PWM]
;  PB6   = EEPROM SDA
;  PB7   = Speaker PWM signal 2 [LPG1PWM]
;
;PMS152: PA5 is 4 mA sink, open drain pin. PB3,5,7 are 30 mA sink pins.
;        All other pins are 14 mA sink. All pins are 7 mA drive (except PA5).
;        => All LED pins are 14 mA sink, 7 mA drive.
;        Measured LED current: ~30 mA at 4.5 V (5 mA average with 1/6 mux).
;
;PFS154: PA5 is 5 mA sink, open drain pin. PA6,7, PB1,3,4,7 are 6 mA sink pins.
;        All other pins are 10 mA sink. All pins are 5 mA drive (except PA5).
;        => LED pins PA6, PA7, PB1 are weaker! Brightness will be uneven!
;        Measured LED current: 30-50 mA at 4.5 V
;
;BUTTON FUNCTION:
;  1. Press button for >1s to turn power on or off.
;  2. Press button for <1s to play sound.
;  3. Hold button while connecting battery to enable always-on mode (button
;     will ignore long presses). Animations will start immediately when button
;     is released. Remove battery to disable always-on mode.
;  4. Hold button while connecting battery and keep holding to enter test mode
;     (for hardware debug: turn on each LED in sequence, repeat indefinitely).
;     Button can be released once test mode starts. Remove battery to exit.
;
;LED ARRANGEMENT
;  See animations/ani_list.asm-pdk
;
;VERSIONS
;  v1  2024-01-26  Arne Rossius
;   * first version
;  v2  2024-01-28  Arne Rossius
;   * avoid lock-up if I2C communication fails during sound playback
;   * added test mode
;  v3  2024-02-24  Arne Rossius
;   * added fix for PCB Rev. 3.0 swapped PA3/PA4

;===============================================================================

;SETTINGS

;button hold-down delay
BUTTON_SHORT_TIME =   2 ;short press (play sound) - 10 ms units
BUTTON_LONG_TIME  = 100 ;long press (power off) - 10 ms units, max. 127
POWER_ON_DELAY    = 250 ;long press (power on) - 4 ms units

;Number of LEDs (max. 36, must be even)
N_LED = 34

;PCB Rev. 3.0 swapped PA3 & PA4 (disable for PCB Rev. 3.1 or later)
PCB_30_FIX = 0

;distance between PWM_6 and PWM_7:    >= 7
;distance between any other 2 values: >= 3

;Gamma 2.2
;PWM_0 = 0 ;don't change
;PWM_1 = 3
;PWM_2 = 13
;PWM_3 = 32
;PWM_4 = 61
;PWM_5 = 99
;PWM_6 = 148
;PWM_7 = 208 ;8 MHz / 64 / 6 / 208 = ~100.16 Hz

PWM_0 = 0 ;don't change
PWM_1 = 3
PWM_2 = 7
PWM_3 = 13
PWM_4 = 26
PWM_5 = 52
PWM_6 = 104
PWM_7 = 208 ;8 MHz / 64 / 6 / 208 = ~100.16 Hz

;Port pins
BUTTON BIT 5 ;PA5
SCL    BIT 0 ;PB0
PWM0   BIT 5 ;PB5
SDA    BIT 6 ;PB6
PWM1   BIT 7 ;PB7

  SEGMENT CODE
  IF (MOMCPUNAME == "PMS152")
    INCLUDE "../include/pms152.inc"
    ORG FUSE_WORD
    DATA FUSE_BITS_HIGH | FUSE_SECURITY_OFF | FUSE_BOOTUP_SLOW
  ELSEIF (MOMCPUNAME == "PFS154")
    INCLUDE "../include/pfs154.inc"
    ORG FUSE_WORD
    DATA FUSE_BITS_HIGH | FUSE_SECURITY_OFF | FUSE_IO_DRV_NORMAL | FUSE_BOOTUP_SLOW
  ELSE
    ERROR "Selected CPU not supported!"
  ENDIF
  INCLUDE "../include/easypdk.inc"

;===============================================================================

;NOTE: RAM address for t0sn/t1sn/set0/set1 must be 6 bit (0x00~0x3F)
	SEGMENT DATA
	ORG 0x00
	
FrameShadow: RES 36/2
	IF (FrameShadow != 0)
	  OUTRADIX 16
	  ERROR "FrameShadow must be at address 0x00, is at 0x\{FrameShadow}."
	ENDIF

	ALIGN 2
Command: RES 2 ;Animation command (read from ROM, 14 bit) & 2nd RAM pointer

	ALIGN 2
IntPtr: RES 1 ;RAM pointer for interrupts

Count: RES 1 ;General-purpose counter
Repeat: RES 1 ;Animation repetition counter
AniPtr: RES 1 ;Animation pointer (ROM pointer, low byte only)
NewDelay: RES 1 ;Durarion for new frame (set in main, read in PWM/Mux interrupt
;                                                     when fUpdateLEDs is set)
IntCnt: RES 1 ;Interrupt counter register (used in interrupts)
LEDs: RES 1 ;LED data (used in PWM/Mux interrupt)
PwmTimer: RES 1 ;PWM timer load value (used in PWM/Mux interrupt)
PWM: RES 1 ;PWM step (used in PWM/Mux interrupt)
Mux: RES 1 ;Mux step (used in PWM/Mux interrupt)
BtnCnt: RES 1 ;Button press duration counter (used in PWM/Mux interrupt)
Delay: RES 1 ;Frame duration counter (used in PWM/Mux interrupt)

FrameLED: RES 36/2
	IF (FrameLED != 0x20)
	  OUTRADIX 16
	  ERROR "FrameLED must be at address 0x20, is at 0x\{FrameLED}."
	ENDIF

FrameBackup: RES 36/2

RAM_Clear_End:

	ALIGN 2
Pointer: RES 1 ;RAM pointer (must be after RAM_Clear_End)

DelayBackup: RES 1 ;backup for 'Delay' during sound playback

	ALIGN 2
Stack: RES 10
	;stack requirement:
	;  3 nested REPEATs (3x 2)
	;  interrupt (2) -or- ROM read address (2, with interrupts disabled)
	;  pushaf (2)

;===============================================================================

Flags SFR GPCC ;comparator control register, abused as flag register (save RAM)
fButtonShort	BIT 0
fButtonLong	BIT 1
fUpdateLEDs	BIT 2
fAlwaysOn	BIT 3 ;always-on mode (can't turn on/off with button)
fNibbleBit	BIT 4
		;bit 5 is read-only
		;bit 7 enables the comparator, should remain 0

;===============================================================================

	SEGMENT CODE
	ORG 0x000 ;reset vector
	
	IF (MOMCPUNAME == "PMS152")
	  mov	a, MISCLVR_2V7_gc | MISCLVR_BG_AUTO_gc
	  mov	MISCLVR, a
	  mov	a, #0
	  mov	ROP, a
	ELSEIF (MOMCPUNAME == "PFS154")
	  mov	a, MISCLVR_2V75_gc | MISCLVR_BG_AUTO_gc
	  mov	MISCLVR, a
	ENDIF
	
	goto start
	
;===============================================================================
	
	ORG INT_VECT ;interrupt vector
	
	;Timer 2 overflow (LED PWM/Multiplexing)
	set0	INTRQ, INTRQ_TM2 ;clear interrupt flag
	pushaf
	
	;set timer period for next interrupt
	;NOTE: Can't write TM2CT when TM2CT == 0 (?!), so using TM2B instead
	mov	a, PwmTimer
	mov	TM2B, a
	
	;LEDs off
	mov	a, 0
	mov	PA, a
	set0	PB, 1
	set0	PB, 2
	mov	PAC, a
	set0	PBC, 1
	set0	PBC, 2
	
	;LEDs on with new pattern
	;set anode pin + active LED pins to output
	mov	a, LEDs
	mov	PAC, a
	and	a, 0x06
	xor	PBC, a
	
	;set anode pin high
	mov	a, Mux
	mov	PA, a
	and	a, 0x06
	xor	PB, a
	
	;IRQ+19 cycles from timer overflow to here
	
	;TODO: Use mux as outer, PWM as inner loop to increase mux frequency
	;TODO: Use random PWM order similar to BlinkenPlus?
	
	;LED PWM
	mov	a, 0x10
	add	PWM, a
	t1sn	PWM, 7 ;PWM >= 0x80?
	goto	tm2_calc ;+5 cycles if this path taken
	;restart PWM sequence
	mov	PWM, a ;valid PWM values: 0x10..0x70, step 0x10
	
	;LED Multiplexing
	sl	Mux ;count 0x01, 0x02, 0x04, 0x08, 0x10, skip, 0x40
	t0sn	Mux, 5
	sl	Mux ;skip PA5
	t1sn	Mux, 7
	goto	tm2_calc ;+11 cycles if this path taken
	;restart multiplexing sequence
	mov	a, 0x01
	mov	Mux, a
	
	;+12 cycles to here
	
	;read button (~100 Hz) - 9 cycles (11 cycles after button release)
	t0sn	PA, BUTTON
	goto	tm2_button_released
	;button pressed
	t1sn	BtnCnt, 7 ;avoid overflow
	inc	BtnCnt
	mov	a, BtnCnt
	cneqsn	a, BUTTON_LONG_TIME
	set1	Flags, fButtonLong
	goto	tm2_button_end
tm2_button_released:
	;button not pressed
	mov	a, BtnCnt
	clear	BtnCnt
	sub	a, BUTTON_SHORT_TIME
	t0sn	FLAG, FLAG_C
	goto	tm2_button_end
	sub	a, BUTTON_LONG_TIME - BUTTON_SHORT_TIME
	t0sn	FLAG, FLAG_C
	set1	Flags, fButtonShort
tm2_button_end:

	;decrement delay and update LEDs (~100 Hz) - max. 189 cycles
	dec	Delay
	addc	Delay ;restore to 0 if 'dec' caused underflow
	t0sn	Flags, fUpdateLEDs
	t1sn	FLAG, FLAG_Z
	goto	tm2_update_end ;6 cycles if this path taken
	set0	Flags, fUpdateLEDs
	mov	a, NewDelay
	mov	Delay, a
	;copy data from FrameShadow to FrameLED
	mov	a, #FrameShadow + N_LED/2 - 1
	mov	IntPtr, a
tm2_update_loop:
	idxm	a, IntPtr
	set1	IntPtr, 5 ;FrameLED starts at 0x20
	idxm	IntPtr, a
	set0	IntPtr, 5 ;FrameShadow starts at 0x00
	dec	IntPtr
	t1sn	FLAG, FLAG_C ;IntPtr < FrameShadow ?
	goto	tm2_update_loop
tm2_update_end:

tm2_calc:
	;calculate new values for LEDs (output in next interrupt)
	;13 + 6*12+4 = 89 cycles
	mov	a, #FrameLED << 1 ;Addr = FrameLED + 6/2*(column-1)
	;(add 0 for column 1 [PA0] = Anode for LEDs 1..6)
	t0sn	Mux, 1 ;Column 2 (PB1) = Anode for LEDs 7..12
	add	a, (3*1) << 1
	t0sn	Mux, 2 ;Column 3 (PB2) = Anode for LEDs 13..18
	add	a, (3*2) << 1
	IF (PCB_30_FIX)
	  t0sn	Mux, 4 ;Column 4 (PA4!) = Anode for LEDs 19..24
	  add	a, (3*3) << 1
	  t0sn	Mux, 3 ;Column 5 (PA3!) = Anode for LEDs 25..30
	  add	a, (3*4) << 1
	ELSE
	  t0sn	Mux, 3 ;Column 4 (PA3) = Anode for LEDs 19..24
	  add	a, (3*3) << 1
	  t0sn	Mux, 4 ;Column 5 (PA4) = Anode for LEDs 25..30
	  add	a, (3*4) << 1
	ENDIF
	t0sn	Mux, 6 ;Column 6 (PA6!) = Anode for LEDs 31..36
	add	a, (3*5) << 1
	;(column 7 [PA7] doesn't exist)
	mov	IntPtr, a
	clear	LEDs
	
; 	mov	a, 0x01 ;count 0x01, 0x02, 0x04, 0x08, 0x10, skip, 0x40, 0x80
; 	mov	IntCnt, a
; tm2_led_loop:
; 	mov	a, IntCnt
; 	cneqsn	a, Mux
; 	goto	tm2_led_on ;IntCnt == Mux: anode pin, always on
; 	sr	IntPtr ;remove nibble bit from address
; 	idxm	a, IntPtr
; 	slc	IntPtr ;restore nibble bit
; 	;t0sn	IntPtr, 0
; 	;swap	a ;odd nibble stored in RAM bits 7:4
; 	;and	a, 0x0F ;mask the nibble
;  	t1sn	IntPtr, 0
;  	swap	a ;move even nibbles from bits 3:0 to bits 7:4
; 	inc	IntPtr
; 	sub	a, PWM
; 	mov	a, IntCnt
; 	t1sn	FLAG, FLAG_C ;C=1: value < PWM => LED off
; tm2_led_on:
; 	or	LEDs, a ;LED on
; 	sl	IntCnt
; 	t0sn	IntCnt, 5
; 	sl	IntCnt ;skip PA5
; 	t1sn	FLAG, FLAG_C
; 	goto	tm2_led_loop
	
LED_PWM MACRO b
	;12 cycles (4 cycles if Mux[b] == 1)
	t0sn	Mux, b
	goto	tm2_led_on
	sr	IntPtr ;remove nibble bit from address
	idxm	a, IntPtr
	slc	IntPtr ;restore nibble bit
	t1sn	IntPtr, 0
	swap	a ;move even nibbles from bits 3:0 to bits 7:4
	inc	IntPtr
	sub	a, PWM
	t1sn	FLAG, FLAG_C ;C=1: value < PWM => LED off
tm2_led_on:
	set1	LEDs, b
	
	ENDM
	
	LED_PWM 0 ;Row 1 = PA0
	LED_PWM 1 ;Row 2 = PB1
	LED_PWM 2 ;Row 3 = PB2
	IF (PCB_30_FIX)
	  LED_PWM 4 ;Row 4 = PA4 (!)
	  LED_PWM 3 ;Row 5 = PA3 (!)
	ELSE
	  LED_PWM 3 ;Row 4 = PA3
	  LED_PWM 4 ;Row 5 = PA4
	ENDIF
	LED_PWM 6 ;Row 6 = PA6 (!)
	LED_PWM 7 ;Row 7 = PA7
	
	;get timer bound for next PWM interval (used in next IRQ) - 6~12 cycles
PWMADD7 = PWM_7-PWM_6 - 1 - 7
PWMADD6 = PWM_6-PWM_5 - 1 - 6 - PWMADD7
PWMADD5 = PWM_5-PWM_4 - 1 - 5 - PWMADD7-PWMADD6
PWMADD4 = PWM_4-PWM_3 - 1 - 4 - PWMADD7-PWMADD6-PWMADD5
PWMADD3 = PWM_3-PWM_2 - 1 - 3 - PWMADD7-PWMADD6-PWMADD5-PWMADD4
PWMADD2 = PWM_2-PWM_1 - 1 - 2 - PWMADD7-PWMADD6-PWMADD5-PWMADD4-PWMADD3
PWMADD1 = PWM_1-PWM_0 - 1 - 1 - PWMADD7-PWMADD6-PWMADD5-PWMADD4-PWMADD3-PWMADD2
;         \_________/   ^   ^   \_____________________________________________/
;       Desired Period  |   |       Values added to 'a' in subsequent lines
;                       |   |
;   Bound = Period-1 ---'   '--- Value in 'a' before 'pcadd'
	mov	a, PWM ;PWM == 0x10..0x70
	swap	a ;a = 1..7
	pcadd	a
	add	a, PWMADD1 ;PWM == 0x10
	add	a, PWMADD2 ;PWM == 0x20
	add	a, PWMADD3 ;PWM == 0x30
	add	a, PWMADD4 ;PWM == 0x40
	add	a, PWMADD5 ;PWM == 0x50
	add	a, PWMADD6 ;PWM == 0x60
	add	a, PWMADD7 ;PWM == 0x70
	mov	PwmTimer, a
	
	;return from interrupt - 4 cycles
	set0	PB, SDA ;See [1] below
	popaf
	reti
	
;TOTAL IRQ TIME IN CYCLES [TICKS], 1 timer tick = 64 cycles
;  Last PWM step, last Mux step, new frame: 19+12+9+189+89+12+4 = 334 [5.22]
;  Last PWM step, last Mux step, same frame: 19+12+9+6+89+12+4 = 151 [2.36]
;  Last PWM step, other Mux step: 19+11+89+12+4 = 135 [2.11]
;  Other PWM step: 19+4+89+12+4 = 128 [2.00]
;Time between interrupts: 3-4 cycles (1-2 for instruction + 2 for jump to IRQ)

;===============================================================================

;animation command macros

FILL MACRO led1, nn, brt ;--00.nnnl:llll.lbbb
	IF ((led1 < 1) || (led1 > N_LED))
	  ERROR "FILL: LED number out of range (1~N_LED)."
	ENDIF
	IF ((nn < 0) || (nn > 6))
	  ERROR "FILL: stretch length out of range (0~6)."
	ENDIF
	IF ((brt < 0) || (brt > 7))
	  ERROR "FILL: Brightness out of range (0~7)."
	ENDIF
	DATA nn << 9 | (led1-1) << 3 | brt
	ENDM

LED MACRO led1, brt
	FILL led1, 0, brt
	ENDM

ALLOFF MACRO
	FILL 1, 6, 0
	ENDM

ROTATE MACRO nn ;--10.0000:00nn.nnnn
	IF ((nn >= 1) && (nn <= N_LED-1))
	  DATA 0x2000 | nn
	ELSEIF ((nn >= -(N_LED-1)) && (nn <= -1))
	  DATA 0x2000 | (N_LED + nn)
	ELSE
	  ERROR "ROTATE: value out of range (-[N_LED-1]~-1 or 1~[N_LED-1])."
	ENDIF
	ENDM

FADE MACRO nn ;--10.0001:nnnn.nnnn
	IF ((nn >= 1) && (nn <= 7))
	  DATA 0x2100 | nn
	ELSEIF ((nn >= -7) && (nn <= -1))
	  DATA 0x2100 | (0x100 + nn)
	ELSE
	  ERROR "FADE: value out of range (-7~-1 or 1~7)."
	ENDIF
	ENDM

DELAY MACRO nn ;--10.0010:nnnn.nnnn
	IF ((nn < 10) || (nn > 255*10))
	  ERROR "DELAY: value out of range (10~2550)."
	ENDIF
	DATA 0x2200 | ((nn + 5) / 10)
	ENDM

REPEAT MACRO nn ;--10.0100:nnnn.nnnn
	IF ((nn < 1) || (nn > 255))
	  ERROR "REPEAT: value out of range (1~255)."
	ENDIF
	DATA 0x2300 | nn
	ENDM

REPEAT_END MACRO
	DATA 0x2300
	ENDM

ANI_END MACRO
 	DATA 0x2400
	ENDM

	
	ALIGN 256
animations:
	INCLUDE "animations/ani_list.pdk.asm"
animations_end:
ANI_SIZE = animations_end - animations
	IF (ANI_SIZE > 256)
	  OUTRADIX 10
	  ERROR "Animation size (\{ANI_SIZE}) exceeds 256 words!"
	ENDIF
	
;===============================================================================
	
start:
	;select fast wake-up
	mov	a, MISC_WAKEUP_FAST_gc
	mov	MISC, a
	
	;set oscillator
	;switch SYSCLK to IHRC/2 (8 MHz), turn off watchdog
	mov	a, CLKMD_IHRC_DIV2_gc | 1<<CLKMD_IHRC_EN | 1<<CLKMD_ILRC_EN
	mov	CLKMD, a
	;turn off ILRC
	set0	CLKMD, CLKMD_ILRC_EN
	
	;clock frequency calibration
	EASYPDK_CALIBRATE_IHRC 8000000, 4500 ;8 MHz @ 4.5 V
	
	;bandgap voltage calibration
	EASYPDK_CALIBRATE_BG
	
	;init ports
	mov	a, 1<<BUTTON ;enable digital input + pull-up on button pin
	mov	PADIER, a
	mov	PAPH, a
	mov	a, 0x00 ;disable all digital inputs on port B
	mov	PBDIER, a
	mov	a, 0x18 | 1<<SCL | 1<<SDA ;enable pull-up on unused & I2C pins
	mov	PBPH, a
	mov	a, 1<<PWM0 | 1<<PWM1 ;PWM pins = output
	mov	PBC, a
	
	;set stack pointer
	mov	a, #Stack
	mov	SP, a
	
	;clear RAM (Frame buffers and variables)
	mov	a, #RAM_Clear_End - 1
	mov	Pointer, a
	mov	a, 0x00
	mov	Flags, a
init_ram:
	idxm	Pointer, a
	dzsn	Pointer
	goto	init_ram
	idxm	Pointer, a
	
	;init non-zero variables
	set1	PwmTimer, 2 ;PwmTimer := 4
	set1	PWM, 7 ;PWM := 0x80
	set1	Mux, 6 ;Mux := 0x40
	
	;read button (determine always-on or test mode)
	clear	Count
init_mode:
	;read button
	t0sn	PA, 5
	goto	init_mode_end ;button not pressed
	mov	a, 0 ;delay ~24.64 ms (256 * [256*3 + 2] = 197120 cycles)
init_mode_delay:
	dzsn	a
	goto	init_mode_delay
	dzsn	Repeat
	goto	init_mode_delay
	set1	Flags, fAlwaysOn ;button held down at power-on: always-on mode
	inc	Count
	mov	a, Count
	ceqsn	a, 203 ;~5 s
	goto	init_mode
	;button held down for 5 s: enter test mode
	goto	test_mode
init_mode_end:
	
	;go to standby
	t1sn	Flags, fAlwaysOn ;always-on mode?
	goto	power_off
	goto	power_on
	
;-------------------------------------------------------------------------------
	
ani_start:
	clear	AniPtr
ani_loop:
	mov	a, SP
	mov	Pointer, a
	mov	a, AniPtr
	disgint
	idxm	Pointer, a ;push ROM read address to stack
	inc	Pointer
	mov	a, #animations >> 8
	idxm	Pointer, a
	ldsptl ;read animation command from ROM (low byte)
	mov	Command, a
	ldspth ;read animation command from ROM (high byte)
	engint
	mov	Command+1, a
	inc	AniPtr ;increment animation ROM pointer
	cneqsn	a, 0x20 ;--10.0000 - rotate
	goto	ani_rotate
	;NOTE: 'cneqsn' sets flags as in 0x20-a (not a-0x20)! Datasheet is wrong!
	t1sn	FLAG, FLAG_C ;--00.xxxx - fill LEDs with value
	goto	ani_fill
	cneqsn	a, 0x21 ;--10.0001 - fade
	goto	ani_fade
	cneqsn	a, 0x22 ;--10.0010 - delay
	goto	ani_delay
	cneqsn	a, 0x23 ;--10.0011 - repeat, repeat end
	goto	ani_repeat
	cneqsn	a, 0x24 ;--10.0100 - end of animations
	goto	ani_start
	;--10.0101 ~ --11.1111 reserved for future extensions
	goto	ani_loop
	
;--------------------
	
ani_delay:
	;delay: new_delay * 10 ms
	mov	a, Command
	mov	NewDelay, a
	set1	Flags, fUpdateLEDs
ani_delay_wait:
	t0sn	Flags, fButtonShort
	goto	sound
	t1sn	Flags, fButtonLong
	goto	ani_delay_wait_end
	t1sn	Flags, fAlwaysOn ;always-on mode?
	goto	power_off
power_off_return:
ani_delay_wait_end:
	t0sn	Flags, fUpdateLEDs
	goto	ani_delay_wait
	goto	ani_loop
	
;--------------------

ani_fade:
	;fade
	mov	a, #FrameShadow + N_LED/2 - 1
	mov	Pointer, a ;FrameShadow
ani_fade_loop:
	idxm	a, Pointer
	mov	Count, a
	;lower nibble
	and	a, 0x0F
	add	a, Command
	mov	Command+1, a
	t0sn	Command+1, 3 ;larger than 7 (max. 7+7 = 14)
	mov	a, 7
	t0sn	Command+1, 7 ;negative?
	mov	a, 0
	;upper nibble
	xch	Count
	swap	a
	and	a, 0x0F
	add	a, Command
	mov	Command+1, a
	t0sn	Command+1, 3 ;larger than 7 (max. 7+7 = 14)
	mov	a, 7
	t0sn	Command+1, 7 ;negative?
	mov	a, 0
	swap	a
	;store
	or	a, Count ;combine nibbles
	idxm	Pointer, a
	dec	Pointer
	t1sn	FLAG, FLAG_C
	goto	ani_fade_loop
	goto	ani_loop
	
;--------------------

ani_fill:
	;--00.llla, aaaa.avvv - fill 2^l LEDs, starting at LED a, with value v
	mov	a, Command
	and	a, 0x07
	xch	Command+1 ;LED value to Command+1, get --00.llla
	sr	a
	xch	Command ;lll to Command, get aaa.avvv
	src	a
	sr	a
	sr	a
	mov	Pointer, a ;FrameShadow == 0
	mov	a, 1
	t0sn	Command, 1
	mov	a, 4
	t0sn	Command, 0
	sl	a
	t0sn	Command, 2
	swap	a
	mov	Count, a ;number of LEDs (2^l) to Count
ani_fill_loop:
	mov	a, Pointer
	cneqsn	a, N_LED ;wrap around for address
	clear	Pointer ;Pointer := FrameShadow
	sr	Pointer ;remove nibble bit
	idxm	a, Pointer
	t0sn	FLAG, FLAG_C
	swap	a ;odd nibble stored in RAM bits 7:4
	and	a, 0xF0 ;clear old value
	or	a, Command+1 ;set new value
	t0sn	FLAG, FLAG_C
	swap	a
	idxm	Pointer, a
	slc	Pointer ;restore nibble bit
	inc	Pointer
	dzsn	Count
	goto	ani_fill_loop
	goto	ani_loop
	
;--------------------

ani_rotate:
	;rotate
	
	;copy LED values to backup framebuffer at shifted address
	mov	a, #FrameShadow + N_LED/2 - 1 ;get high nibble of last byte
	mov	Pointer, a
	idxm	a, Pointer
	swap	a
	and	a, 0x0F
	mov	Count, a
	clear	Pointer ;FrameShadow
	sr	Command ;remove nibble bit ...
	swapc	Flags, fNibbleBit ;... and store it to Flags
	mov	a, #FrameBackup
	add	Command, a ;FrameBackup + offset
ani_rotate_loop1:
	idxm	a, Pointer
	inc	Pointer
	t1sn	Flags, fNibbleBit
	goto	ani_rotate_store
	;shift by an additional nibble
	swap	a
	mov	Command+1, a
	and	a, 0xF0 ;current byte low nibble -> current byte high nibble
	xch	Command+1
	and	a, 0x0F ;current byte high nibble -> next byte low nibble,
	xch	Count   ;previous byte high nibble -> current byte low nibble
	or	a, Command+1
ani_rotate_store:
	idxm	Command, a
	inc	Command
	mov	a, Command
	cneqsn	a, #FrameBackup + N_LED/2 ;wrap around
	mov	a, #FrameBackup
	mov	Command, a
	mov	a, Pointer
	ceqsn	a, #FrameShadow + N_LED/2 ;stop at end of frame
	goto	ani_rotate_loop1
	
	;copy shifted values back to displayed framebuffer
	mov	a, #FrameBackup + N_LED/2
	mov	Command, a
	;Pointer == FrameShadow + N_LED/2 (from previous loop)
ani_rotate_loop2:
	dec	Command
	dec	Pointer
	idxm	a, Command
	idxm	Pointer, a
	t1sn	FLAG, FLAG_Z ;stop when Pointer == 0 (== FrameShadow)
	goto	ani_rotate_loop2
	
	goto	ani_loop
	
;--------------------

ani_repeat:
	;repeat
	mov	a, Command
	cneqsn	a, 0
	goto	ani_repeat_end ;"REPEAT 0" = end of repeat loop
	mov	a, Repeat ;store repeat counter (for nested repeats)
	pushaf
	mov	a, SP
	mov	Pointer, a
	dec	Pointer
	mov	a, AniPtr ;store start-of-repeat-block address
	idxm	Pointer, a
	mov	a, Command ;load new repeat counter
	mov	Repeat, a
	goto	ani_loop
	
;--------------------

ani_repeat_end:
	;repeat end
	dzsn	Repeat
	goto	ani_repeat_reload
	
	;repeat loop complete
	mov	a, SP
	sub	a, 2
	mov	Pointer, a
	idxm	a, Pointer ;restore repeat counter
	mov	Repeat, a
	popaf ;free stack
	goto	ani_loop

ani_repeat_reload:
	;restart repeat loop
	mov	a, SP ;read start-of-repeat-block address from stack
	mov	Pointer, a
	dec	Pointer
	idxm	a, Pointer
	mov	AniPtr, a
	
	goto	ani_loop
	
;--------------------

power_off:
	;disable mux timer
	mov	a, 0
	mov	TM2C, a ;disable timer
	disgint ;disable interrupts
	
	;LEDs off (force LED pins high to avoid leakage through transistors)
	mov	a, 0xD9 ;LEDs on PA7, PA6, PA4, PA3, PA0
	mov	PA, a
	mov	a, 0x06 ;LEDs on PB2, PB1
	mov	PB, a
	mov	a, 0xD9
	mov	PAC, a
	mov	a, 1<<PWM0 | 1<<PWM1 | 0x06
	mov	PBC, a
	
	;wait until button released
power_off_wait:
	t1sn	PA, BUTTON
	goto	power_off_wait
	
power_off_sleep:
	;sleep (wake up when any pin enabled in PADIER/PBDIER toggles)
	stopsys
	
	;check if button pressed for a certain time (filter ESD and spikes)
	clear	Count
power_on_loop:
	;delay ~4 ms (~32k cycles / 8 MHz)
	mov	a, 42
	mov	BtnCnt, a
	mov	a, 141
power_on_delay_loop:
	dzsn	a
	goto	power_on_delay_loop
	dzsn	BtnCnt
	goto	power_on_delay_loop
	;poll button
	t0sn	PA, BUTTON
	goto	power_off_sleep
	;button pressed
	inc	Count
	mov	a, Count
	ceqsn	a, POWER_ON_DELAY
	goto	power_on_loop

power_on:
	;set Timer2 for muxing
	mov	a, 255 ;upper bound
	mov	TM2B, a
	mov	a, TM2S_PRESCALE_DIV64_gc | 0<<TM2S_SCALE ;Clk/64/1 = 125 kHz
	mov	TM2S, a
	mov	a, TM2C_CLK_SYSCLK_gc ;Clock = SYSCLK (8 MHz), enable
	mov	TM2C, a
	set0	INTRQ, INTRQ_TM2 ;clear interrupt flag
	set1	INTEN, INTEN_TM2 ;enable TM2 interrupt
	engint ;enable interrupts
	
	mov	a, BUTTON_LONG_TIME + 1 ;avoid button press detection
	mov	BtnCnt, a
	set0	Flags, fButtonShort
	set0	Flags, fButtonLong
	goto	power_off_return

;--------------------

sound:
	;stop further display updates
	mov	a, Delay
	mov	DelayBackup, a
	set0	Flags, fUpdateLEDs
	clear	Delay
	
	;init PWM generator
	mov	a, (0x80 << 5) & 0xFF ;initial duty cycle = 0x80 (50%)
	mov	PWMG0DTL, a
	mov	PWMG1DTL, a
	mov	a, (0x80 >> 3)
	mov	PWMG0DTH, a
	mov	PWMG1DTH, a
	IF (MOMCPUNAME == "PMS152")
	  mov	a, (255 << 5) & 0xFF ;8-bit PWM (0~255) => 8MHz/256 = 31250 Hz
	  mov	PWMGCUBL, a
	  mov	a, (255 >> 3)
	  mov	PWMGCUBH, a
	  mov	a, PWMG0_PIN_PB5_gc ;PWM generator 0 output on PB5
	  mov	PWMG0C, a
	  mov	a, PWMG1_PIN_PB7_gc ;PWM generator 1 output on PB7
	  mov	PWMG1C, a
	  ;NOTE: PGMG_PRESCALE_DIV2 means SYSCLK/1 (i.e. 8 MHz) !!!
	  mov	a, 1<<PWMG_ENABLE | PWMG_PRESCALE_DIV2_gc | PWMG_CLK_SYSCLK_gc
	  mov	PWMGCLK, a ;start PWM generator
	ELSEIF (MOMCPUNAME == "PFS154")
	  mov	a, (255 << 5) & 0xFF ;8-bit PWM (0~255) => 8MHz/256 = 31250 Hz
	  mov	PWMG0CUBL, a
	  mov	PWMG1CUBL, a
	  mov	a, (255 >> 3)
	  mov	PWMG0CUBH, a
	  mov	PWMG1CUBH, a
	  ;NOTE: PWMGx_SCALE in steps of 0.5 (0=0.5, 1=1.0, 2=1.5, 3=2.0, ...)
	  mov	a, PWMG0_INTMODE_ZERO_gc | PWMG0_PRESCALE_DIV1_gc | 1<<PWMG0_SCALE
	  mov	PWMG0S, a
	  mov	a, PWMG1_INTMODE_ZERO_gc | PWMG1_PRESCALE_DIV1_gc | 1<<PWMG1_SCALE
	  mov	PWMG1S, a
	  mov	a, 1<<PWMG0_ENABLE | PWMG0_PIN_PB5_gc | PWMG0_CLK_SYSCLK_gc
	  mov	PWMG0C, a
	  mov	a, 1<<PWMG1_ENABLE | PWMG1_PIN_PB7_gc | PWMG1_CLK_SYSCLK_gc
	  mov	PWMG1C, a
	  ;synchronize PWM generators
	  mov	a, 0x60
pwm_sync1:
	  ceqsn	a, PWM
	  goto	pwm_sync1 ;wait for PWM = 6
	  mov	a, 0x70
pwm_sync2:
	  ceqsn	a, PWM
	  goto	pwm_sync2 ;wait for PWM = 7 (long time until next interrupt)
	  set1	PWMG0C, PWMG0_RESET ;reset PWM0
	  mov	a, 85
pwm_sync_delay:
	  dzsn	a
	  goto	pwm_sync_delay
	  set1	PWMG1C, PWMG1_RESET ;reset PWM1 exactly 256 cycles after PWM1
	ENDIF
	
	;backup framebuffer
	mov	a, #FrameShadow + N_LED/2 - 1
	mov	Pointer, a
	mov	a, #FrameBackup + N_LED/2 - 1
	mov	Command, a
	;clear	Command+1
sound_leds_backup:
	idxm	a, Pointer
	idxm	Command, a
	dec	Command
	dec	Pointer
	t1sn	FLAG, FLAG_C ;FrameShadow == 0
	goto	sound_leds_backup

	;fade all LEDs on
fade_on:
	clear	Command ;LED value := 0
fade_on_loop1:
	mov	a, #FrameShadow + N_LED/2 - 1
	mov	Pointer, a
fade_on_loop2:
	idxm	a, Pointer
	mov	Command+1, a
	;low nibble
	and	a, 0x0F
	sub	a, Command
	add	a, Command
	t0sn	FLAG, FLAG_C
	mov	a, Command
	;high nibble
	xch	Command+1
	swap	a
	and	a, 0x0F
	sub	a, Command
	add	a, Command
	t0sn	FLAG, FLAG_C
	mov	a, Command
	swap	a
	;store
	or	a, Command+1 ;combine nibbles
	idxm	Pointer, a
	dec	Pointer
	t1sn	FLAG, FLAG_C ;FrameShadow == 0
	goto	fade_on_loop2
	;20 ms delay
	mov	a, 2
	mov	NewDelay, a
	set1	Flags, fUpdateLEDs
fade_on_wait:
	t0sn	Flags, fUpdateLEDs
	goto	fade_on_wait
	;next brightness step
	inc	Command ;increase LED value
	t1sn	Command, 3 ;LED value > 7?
	goto	fade_on_loop1
	
	;delay until LEDs updated
	mov	a, 0
fade_on_finish:
	ceqsn	a, Delay
	goto	fade_on_finish

	;send Address = 0x0000 to EEPROM
	set1	PBC, SDA ;data low (start condition)
	mov	a, 3*9 ;3 bytes: I2C address + 2x 0x00 (EEPROM address := 0)
	set1	PBC, SCL ;clock low
	mov	Count, a
	mov	a, ~0xA0 & 0xFF ;bits are inverted (PBC bit low = pin high)
i2c_write_loop:
	sl	a
	swapc	PBC, SDA ;output data
	or	a, 0x01 ;shift in '1' bits => SDA low for bits 9~27
	set1	PB, SCL ;force clock high
	goto	*+1 ;2 cycles delay
	set0	PB, SCL ;clock low
	dzsn	Count
	goto	i2c_write_loop
	goto	*+1
	goto	*+1
	set1	PB, SCL ;clock high
	goto	*+1
	set0	PBC, SDA ;data high (stop condition)
	
	;short delay between I2C stop and start condition
	mov	a, 3
i2c_delay:
	dzsn	a
	goto	i2c_delay
	
	;start playback: read data from EEPROM
	set1	PBC, SDA ;data low (start condition)
	set1	PBDIER, SDA ;enable digital input for SDA
	;NOTE: set0/1 of any PB (or PBC?) bit seems to load other PB bits with
	;      the actual pin values (i.e. input high will set PB bit high).
	;      Workaround is to set PBDIER bit to zero (PB bit will then
	;      be set to 0) or set0 PB bit immediately before set1 PBC bit. [1]
	;https://www.eevblog.com/forum/blog/eevblog-1144-padauk-programmer-reverse-engineering/msg4606273/#msg4606273
	mov	a, 9 ;send I2C address: 8 bits + ACK
	set0	PB, SCL ;clock low
	mov	Count, a
	mov	a, ~0xA1 & 0xFF ;bits are inverted (PBC bit low = pin high)
	mov	Command, a
i2c_read_addr_loop:
	sl	Command ;shift in '0' bits => release SDA ("high") for 9th bit
	set0	PB, SDA
	swapc	PBC, SDA ;output data
	nop
	set1	PB, SCL ;force clock high
	nop
	mov	a, PB ;read data
	set0	PB, SCL ;clock low
	dzsn	Count
	goto	i2c_read_addr_loop
	;check for I2C errors
	mov	Command, a
	t0sn	Command, SDA
	goto	i2c_error_nack ;no ACK received (SDA was high during 9th bit)
	t1sn	PB, SDA
	goto	i2c_error_low ;SDA stuck low
	
	;stream EEPROM data to PWM generator - min. 108 cycles per loop
	clear	NewDelay ;last received byte value
	clear	Count ;counter for number of identical bytes
	IF (MOMCPUNAME == "PMS152")
	  set0	INTRQ, INTRQ_PWMG ;clear PWM interrupt flag
	ELSEIF (MOMCPUNAME == "PFS154")
	  set0	INTRQ, INTRQ_PWMG0 ;clear PWM interrupt flag
	ENDIF
sound_loop:
	;read one byte from EEPROM
	mov	a, 8
i2c_read_loop:
	set1	PB, SCL ;force clock high
	nop
	swapc	PB, SDA ;read data
	set0	PB, SCL ;clock low
	slc	Command
	dzsn	a
	goto	i2c_read_loop
	mov	a, Command
	set0	PB, SDA
	ceqsn	a, 0x00 ;0x00 = end of data - don't send ACK (stop reading)
	set1	PBC, SDA ;data low (ACK)
	neg	a ;inverse value for 2nd PWM channel
	set1	PB, SCL ;force clock high
	sl	a ;bit position: 6543.210-
	addc	a ;bit position: 6543.2107
	swap	a ;bit position: 2107.6543
	set0	PB, SCL ;clock low
	mov	Command+1, a
	set0	PBC, SDA ;data high (for next receive)
	mov	a, Command
	cneqsn	a, 0x00 ;0x00 = end of data
	goto	sound_end
	;check for too many identical bytes (read error)
	cneqsn	a, NewDelay
	goto	i2c_check_equal
	clear	Count ;not equal to previous byte: reset counter to zero
	mov	NewDelay, a
	goto	i2c_check_end
i2c_check_equal:
	inc	Count ;equal to previous byte: increment counter
	t0sn	FLAG, FLAG_C
	goto	i2c_error_repeat ;read 256 identical bytes => abort playback
i2c_check_end:
	;calculate PWM values (continued)
	sl	a ;bit position: 6543.210-
	addc	a ;bit position: 6543.2107
	swap	a ;bit position: 2107.6543
	mov	Command, a
	and	a, 0xE0 ;bits 210-.----
	;wait for PWM generator interrupt
sound_wait:
	IF (MOMCPUNAME == "PMS152")
	  t1sn	INTRQ, INTRQ_PWMG
	  goto	sound_wait
	  set0	INTRQ, INTRQ_PWMG ;clear PWM interrupt flag
	ELSEIF (MOMCPUNAME == "PFS154")
	  t1sn	INTRQ, INTRQ_PWMG0
	  goto	sound_wait
	  set0	INTRQ, INTRQ_PWMG0 ;clear PWM interrupt flag
	ENDIF
	;set new PWM values
	mov	PWMG0DTL, a
	mov	a, Command
	and	a, 0x1F ;bits ---7.6543
	mov	PWMG0DTH, a
	mov	a, Command+1
	and	a, 0xE0 ;bits 210-.----
	mov	PWMG1DTL, a
	mov	a, Command+1
	and	a, 0x1F ;bits ---7.6543
	mov	PWMG1DTH, a
	goto	sound_loop
	
sound_end:
	set0	PBDIER, SDA ;disable digital input for SDA
	set0	PB, SDA
	
	;send i2c stop
	set1	PBC, SDA ;data low
	goto	*+1
	set1	PB, SCL ;clock high
	goto	*+1
	set0	PBC, SDA ;data high
	set0	PBC, SCL ;stop driving clock (internal pull-up only)
	
	;fade all LEDs back to original values
	mov	a, 7 ;LED value := 7
	mov	Command, a
fade_off_loop1:
	mov	a, #FrameShadow + N_LED/2 - 1
	mov	Pointer, a
fade_off_loop2:
	mov	a, #FrameBackup - FrameShadow
	add	Pointer, a
	idxm	a, Pointer
	mov	Command+1, a
	mov	a, #FrameBackup - FrameShadow
	sub	Pointer, a
	;low nibble
	mov	a, Command+1
	and	a, 0x0F
	sub	a, Command
	add	a, Command
	t0sn	FLAG, FLAG_C
	mov	a, Command
	;high nibble
	xch	Command+1
	swap	a
	and	a, 0x0F
	sub	a, Command
	add	a, Command
	t0sn	FLAG, FLAG_C
	mov	a, Command
	swap	a
	;store
	or	a, Command+1 ;combine nibbles
	idxm	Pointer, a
	dec	Pointer
	t1sn	FLAG, FLAG_C ;FrameShadow == 0
	goto	fade_off_loop2
	;20 ms delay
	mov	a, 2
	mov	NewDelay, a
	set1	Flags, fUpdateLEDs
fade_off_wait:
	t0sn	Flags, fUpdateLEDs
	goto	fade_off_wait
	;next brightness step
	dec	Command ;decrease LED value
	t1sn	FLAG, FLAG_C ;LED value < 0?
	goto	fade_off_loop1
	
	;disable audio PWM
	mov	a, 0
	IF (MOMCPUNAME == "PMS152")
	  mov	PWMGCLK, a ;disable PWM
	ENDIF
	mov	PWMG0C, a ;disconnect PWM0 output (and disable PWM0 on PFS154)
	mov	PWMG1C, a ;disconnect PWM1 output (and disable PWM1 on PFS154)
	
	;re-enable display updates
	mov	a, DelayBackup
	mov	Delay, a
	set1	Flags, fUpdateLEDs

	;return
	set0	Flags, fButtonShort
	goto	ani_delay_wait_end
	
;--------------------

i2c_error_repeat:
	;I2C error: 256 identical bytes read (read error)
	mov	a, (0x80 << 5) & 0xFF ;set PWM back to 0x80
	mov	PWMG0DTL, a
	mov	PWMG1DTL, a
	mov	a, (0x80 >> 3)
	mov	PWMG0DTH, a
	mov	PWMG1DTH, a
	mov	a, 6 ;flash 3 times
	goto	i2c_error_flash
	
i2c_error_nack:
	;I2C error: no ACK received (SDA was high)
	mov	a, 4 ;flash twice
	goto	i2c_error_flash
	
i2c_error_low:
	;I2C error: SDA line stuck low
	mov	a, 2 ;flash once
i2c_error_flash:
	;Toggle LEDs off or on 'a' times to indicate error
	mov	Command, a
	mov	a, 25 ;250 ms delay before starting to flash
	mov	Delay, a
	mov	a, 0x77 ;LEDs are currently on
toggle_leds:
	xor	a, 0x77 ;toggle between both nibbles on (0x77) and off (0x00)
	mov	Pointer, a
	mov	a, 25 ;250 ms frame duration
	mov	NewDelay, a
	mov	a, #FrameShadow + N_LED/2 - 1
	xch	Pointer
toggle_leds_loop:
	idxm	Pointer, a
	dec	Pointer
	t1sn	FLAG, FLAG_C ;FrameShadow == 0
	goto	toggle_leds_loop
	set1	Flags, fUpdateLEDs
toggle_leds_wait:
	t0sn	Flags, fUpdateLEDs
	goto	toggle_leds_wait
	dzsn	Command
	goto	toggle_leds
	
	goto	sound_end
	
;===============================================================================
	
test_led MACRO ANODE, CATHODE
	IF ((ANODE == 1) || (ANODE == 2))
	  set1	PB, ANODE
	  set1	PBC, ANODE
	ELSE
	  set1	PA, ANODE
	  set1	PAC, ANODE
	ENDIF
	IF ((CATHODE == 1) || (CATHODE == 2))
	  set0	PB, CATHODE
	  set1	PBC, CATHODE
	ELSE
	  set0	PA, CATHODE
	  set1	PAC, CATHODE
	ENDIF
	call	test_delay
	IF ((ANODE == 1) || (ANODE == 2))
	  set0	PBC, ANODE
	ELSE
	  set0	PAC, ANODE
	ENDIF
	IF ((CATHODE == 1) || (CATHODE == 2))
	  set0	PBC, CATHODE
	ELSE
	  set0	PAC, CATHODE
	ENDIF
	ENDM

test_mode:
	;test mode for hardware debugging:
	;turn on LEDs one at a time in sequence
	
	mov	a, 0x01 ;select first anode
	mov	Mux, a

test_anode_loop:
	mov	a, 0x01 ;select first cathode
	mov	LEDs, a
	
test_cathode_loop:
	;skip cathode == anode
	mov	a, LEDs
	cneqsn	a, Mux
	sl	LEDs
	t0sn	LEDs, 5 ;skip PA5
	sl	LEDs
	
	IF (PCB_30_FIX)
	  ;swap PA3 & PA4
	  mov	a, LEDs
	  and	a, ~0x18
	  t0sn	LEDs, 3
	  or	a, 0x10
	  t0sn	LEDs, 4
	  or	a, 0x08
	  mov	Command, a
	  mov	a, Mux
	  and	a, ~0x18
	  t0sn	Mux, 3
	  or	a, 0x10
	  t0sn	Mux, 4
	  or	a, 0x08
	  mov	Command+1, a
	ENDIF
	
	mov	a, 250 ;250 * 2 ms = 500 ms per LED
	mov	Repeat, a
test_loop:

	;LEDs off
	mov	a, 0
	mov	PA, a
	set0	PB, 1
	set0	PB, 2
	mov	PAC, a
	set0	PBC, 1
	set0	PBC, 2
	
	;delay 5/6 * ~2 ms = 13333 cycles ([80*3 + 2] + 17*[256*3 + 2] = 13332)
	mov	a, 18
	mov	Count, a
	mov	a, 80
test_delay_off:
	dzsn	a
	goto	test_delay_off
	dzsn	Count
	goto	test_delay_off
	
	;LEDs on: anode & cathode = output, anode = high
	IF (PCB_30_FIX)
	  mov	a, Command ;cathode pin
	  or	a, Command+1 ;anode pin
	ELSE
	  mov	a, LEDs ;cathode pin
	  or	a, Mux ;anode pin
	ENDIF
	mov	PAC, a
	and	a, 0x06
	xor	PBC, a
	IF (PCB_30_FIX)
	  mov	a, Command+1
	ELSE
	  mov	a, Mux
	ENDIF
	mov	PA, a
	and	a, 0x06
	xor	PB, a
	
	;delay 1/6 * ~2 ms = 2667 cycles ([118*3 + 2] + 3*[256*3 + 2] = 2666)
	mov	a, 4
	mov	Count, a
	mov	a, 118
test_delay_on:
	dzsn	a
	goto	test_delay_on
	dzsn	Count
	goto	test_delay_on
	
	dzsn	Repeat
	goto	test_loop
	
	;next cathode
	sl	LEDs
	t0sn	LEDs, 5 ;skip PA5
	sl	LEDs
	t1sn	FLAG, FLAG_C
	goto	test_cathode_loop
	
	;next anode
	sl	Mux
	t0sn	Mux, 5 ;skip PA5
	sl	Mux
	t1sn	Mux, 7 ;no anode on PA7
	goto	test_anode_loop
	
	;repeat indefinitely
	goto	test_mode
	
;===============================================================================

	IF (MOMPASS == 2)
	  OUTRADIX 10
	
	  MESSAGE "Animation size: \{ANI_SIZE} words."
	  MESSAGE "ROM use: \{*} words."
	
	  SEGMENT DATA
	  MESSAGE "RAM use: \{*} bytes."
	ENDIF
	
	IF (PCB_30_FIX)
	  WARNING "Fix for PCB Rev. 3.0 enabled!"
	ENDIF
	